/**
 * @file FigureUtil.cpp
 * @brief LN^tBMA֘A[eBeB֐
 * @author cherub
 */
#include "stdafx.h"
#include <cmath>
#include <gsl/gsl_eigen.h>
#include "FigureUtil.h"
#include "Matrix.h"
#include "Statistics.h"

/**
 * @fn float calc_pose_distance(const std::vector<D3DXVECTOR3> &pc0, const std::vector<D3DXVECTOR3> &pc1)
 * @brief 2̊֐߈ʒu_Qf[^̃[NbȟvZ
 * @param pc0 [in] _Q0
 * @param pc1 [in] _Q1
 * @return [Nbh
 */
float calc_pose_distance(const std::vector<D3DXVECTOR3> &pc0, const std::vector<D3DXVECTOR3> &pc1)
{
	D3DXMATRIX trans = NFigureUtil::CalcTransformBetweenFigures(pc0, pc1);

	D3DXVECTOR3 diff;
	float dist =0.0f;
	for (size_t j = 0;j < pc0.size();++j)
	{
		D3DXVec3TransformCoord(&diff, &pc1[j], &trans);
		diff = pc0[j] - diff;
		dist += D3DXVec3LengthSq(&diff);
	}

	return sqrtf(dist / static_cast<float>(pc0.size()));
}

/**
 * @fn D3DXMATRIX NFigureUtil::CalcTransformBetweenFigures(const std::vector<D3DXVECTOR3> &pc0, const std::vector<D3DXVECTOR3> &pc1)
 * @brief 2̓_Q̊Ԃ̃[NbhaŏgXtH[̌vZ: min dist(pc0, trans(pc1))
 * @param pc0 [in] _Q0
 * @param pc1 [in] _Q1
 * @return gXtH[s
 */
D3DXMATRIX NFigureUtil::CalcTransformBetweenFigures(const std::vector<D3DXVECTOR3> &pc0, const std::vector<D3DXVECTOR3> &pc1)
{
	float top_left = 0, top_right = 0, bottom_left = 0, bottom_right = 0;
	float ax1 = 0, ax2 = 0, az1 = 0, az2 = 0;
	float tx, tz, ty = 0, ry;
	const float fnum = static_cast<float>(pc0.size());

	for (size_t i = 0;i < pc0.size();++i)
	{
		top_left += pc0[i].x * pc1[i].z - pc1[i].x * pc0[i].z;
		bottom_left += pc0[i].x * pc1[i].x + pc0[i].z * pc1[i].z;

		ax1 += pc0[i].x;
		az1 += pc0[i].z;
		ax2 += pc1[i].x;
		az2 += pc1[i].z;

		ty += pc0[i].y - pc1[i].y;
	}

	top_right = (ax1 * az2 - ax2 * az1) / fnum;
	bottom_right = (ax1 * ax2 + az1 * az2) / fnum;

	ry = std::atan2f(top_left - top_right, bottom_left - bottom_right);
	tx = (ax1 - ax2 * cosf(ry) - az2 * sinf(ry)) / fnum;
	tz = (az1 + ax2 * sinf(ry) - az2 * cosf(ry)) / fnum;
	ty /= fnum;

	D3DXMATRIX mr, mt;
	D3DXMatrixRotationY(&mr, ry);
	D3DXMatrixTranslation(&mt, tx, ty, tz);
	return mr * mt;
}

/**
 * @fn D3DXQUATERNION NFigureUtil::CalcRotationBetweenPointClouds(const std::vector<D3DXVECTOR3> &pc0, const std::vector<D3DXVECTOR3> &pc1)
 * @brief 2̓_Q̊Ԃ̃[Nbhaŏ]gXtH[̌vZ min dist(pc0, rot(pc1))
 * @note B.K.P. Horn, "Closed-form solution of absolute orientation using unit quaternions", J. of Optical Society of America, vol.4, pp.629-642, 1987.
 * @param pc0 [in] _Q0
 * @param pc1 [in] _Q1
 * @return ]gXtH[s
 */
D3DXQUATERNION NFigureUtil::CalcRotationBetweenPointClouds(const std::vector<D3DXVECTOR3> &pc0, const std::vector<D3DXVECTOR3> &pc1)
{
	const float fnum = static_cast<float>(pc0.size());

	D3DXVECTOR3 cent0(0, 0, 0), cent1(0, 0, 0);
	for (size_t i = 0;i < pc0.size();++i)
	{
		cent0 += pc0[i];
		cent1 += pc1[i];
	}
	cent0 /= fnum;
	cent1 /= fnum;

	float sxx =0, sxy =0, sxz =0, syx =0, syy =0, syz =0, szx =0, szy =0, szz =0;
	for (size_t i = 0;i < pc0.size();++i)
	{
#if 1 //centering
		sxx += (pc1[i].x - cent1.x) * (pc0[i].x - cent0.x);
		sxy += (pc1[i].x - cent1.x) * (pc0[i].y - cent0.y);
		sxz += (pc1[i].x - cent1.x) * (pc0[i].z - cent0.z);
		syx += (pc1[i].y - cent1.y) * (pc0[i].x - cent0.x);
		syy += (pc1[i].y - cent1.y) * (pc0[i].y - cent0.y);
		syz += (pc1[i].y - cent1.y) * (pc0[i].z - cent0.z);
		szx += (pc1[i].z - cent1.z) * (pc0[i].x - cent0.x);
		szy += (pc1[i].z - cent1.z) * (pc0[i].y - cent0.y);
		szz += (pc1[i].z - cent1.z) * (pc0[i].z - cent0.z);
#else
		sxx += pc1[i].x * pc0[i].x;
		sxy += pc1[i].x * pc0[i].y;
		sxz += pc1[i].x * pc0[i].z;
		syx += pc1[i].y * pc0[i].x;
		syy += pc1[i].y * pc0[i].y;
		syz += pc1[i].y * pc0[i].z;
		szx += pc1[i].z * pc0[i].x;
		szy += pc1[i].z * pc0[i].y;
		szz += pc1[i].z * pc0[i].z;
#endif
	}

	CMatrix mN(4, 4);
	mN(0, 0) = sxx + syy + szz;
	mN(0, 1) = syz - szy;		mN(1, 0) = mN(0, 1);
	mN(0, 2) = szx - sxz;		mN(2, 0) = mN(0, 2);
	mN(0, 3) = sxy - syx;		mN(3, 0) = mN(0, 3);
	mN(1, 1) = sxx - syy - szz;
	mN(1, 2) = sxy + syx;		mN(2, 1) = mN(1, 2);
	mN(1, 3) = szx + sxz;		mN(3, 1) = mN(1, 3);
	mN(2, 2) = -sxx + syy - szz;
	mN(2, 3) = syz + szy;		mN(3, 2) = mN(2, 3);
	mN(3, 3) = -sxx - syy + szz;

	CMatrix evecs;
	CVector evals;
	NStatistics::EVD(evecs, evals, mN);
	// ŗLl̐ΒlŃ\[gĂ̂ŁCt̒lɂčă\[g
	gsl_eigen_symmv_sort(evals.m_pvVector, evecs.m_pmMatrix, GSL_EIGEN_SORT_VAL_DESC);

	D3DXQUATERNION nq, q(evecs(1, 0), evecs(2, 0), evecs(3, 0), evecs(0, 0));
	return *D3DXQuaternionNormalize(&nq, &q);
}

/**
 * @fn float NFigureUtil::PoseDistance(const CFigure &fig0, const CFigure &fig1)
 * @brief 2̃tBMA̎px 
 * @param fig0 [in] tBMA0
 * @param fig1 [in] tBMA1
 * @return px
 */
float NFigureUtil::PoseDistance(const CFigure &fig0, const CFigure &fig1)
{
	std::vector<D3DXVECTOR3> pc0 = fig0.GetPointCloud();
	std::vector<D3DXVECTOR3> pc1 = fig1.GetPointCloud();
	return calc_pose_distance(pc0, pc1);
}

/**
 * @fn float NFigureUtil::PoseDistance(CFigure &fig, const CMotionData &mot, size_t frm0, size_t frm1)
 * @brief 2̃t[ł̎px
 * @param fig [in/out] tBMA
 * @param mot [in] [Vf[^
 * @param frm0 [in] t[0
 * @param frm1 [in] t[1
 * @return px
 */
float NFigureUtil::PoseDistance(CFigure &fig, const CMotionData &mot, size_t frm0, size_t frm1)
{
	fig.SetPose(&mot, frm0);
	std::vector<D3DXVECTOR3> pc0 = fig.GetPointCloud();
	fig.SetPose(&mot, frm1);
	std::vector<D3DXVECTOR3> pc1 = fig.GetPointCloud();
	return calc_pose_distance(pc0, pc1);
}

/**
 * @fn float NFigureUtil::PoseDistanceLowerBody(const CFigure &fig0, const CFigure &fig1)
 * @brief 2̃tBMẢĝ݂̎px
 * @param fig0 [in] tBMA0
 * @param fig1 [in] tBMA1
 * @return px
 */
float NFigureUtil::PoseDistanceLowerBody(const CFigure &fig0, const CFigure &fig1)
{
	std::vector<D3DXVECTOR3> pc0 = fig0.GetPointCloudLowerBody();
	std::vector<D3DXVECTOR3> pc1 = fig1.GetPointCloudLowerBody();
	return calc_pose_distance(pc0, pc1);
}
