/**
 * @file MotionData.h
 * @brief [Vf[^NX錾
 * @author cherub
 */
#ifndef MOTION_DATA_CLASS_HEADER_
#define MOTION_DATA_CLASS_HEADER__
#pragma once

#include <d3dx9math.h>

/**
 * @class CMotionData MotionData.h
 * @brief [Vf[^NX
 * @author cherub
 */
class CMotionData
{
//
// variables
private:
	//! ֐ߐ
	size_t m_nJoints;
	//! t[
	size_t m_nFrames;
	//! [gʒunf[^
	D3DXVECTOR3 *m_pPosition;
	//! ֐߉]pxnf[^
	D3DXQUATERNION *m_pRotation;

//
// constructor/destructor
public:
	// ftHgRXgN^
	CMotionData(void);
	// qtRXgN^
	CMotionData(size_t frames, size_t joints);
	// Rs[RXgN^
	CMotionData(const CMotionData &src);
	// fXgN^
	virtual ~CMotionData(void);

//
// initializer
public:
	// ֐
	bool Initialize(size_t frames, size_t joints);
	// Zq
	CMotionData& operator =(const CMotionData &src);

//
// memory allocator
private:
	// ̈m
	bool Allocate(size_t frames, size_t joints);
	// ̈
	void Destroy();

//
// attributes
public:
	// f[^p\
	bool IsActive() const;
	/**
	 * @fn CMotionData::NumFrames() const
	 * @brief t[̎擾
	 * @return t[
	 */
	size_t NumFrames() const
		{
			return m_nFrames;
		}
	/**
	 * @fn CMotionData::NumJoints() const
	 * @brief ֐߃m[h̎擾
	 * @return ֐߃m[h
	 */
	size_t NumJoints() const
		{
			return m_nJoints;
		}

//
// accessor
public:
	// ]ʂ̐ݒ
	bool SetRotation(size_t frm, size_t jid, const D3DXQUATERNION &q);
	// ]ʂ̐ݒ
	bool SetRotation(size_t frm, size_t jid, const D3DXMATRIX &m);
	// [gʒu̎w
	bool SetPosition(size_t frm, const D3DXVECTOR3 &v);

	// ]ʂ̎擾
	D3DXQUATERNION GetRotation(size_t frm, size_t jid) const;
	// ]ʂ̎擾
	D3DXMATRIX GetRotationMatrix(size_t frm, size_t jid) const;
	// [gʒu̎擾
	D3DXVECTOR3 GetPosition(size_t frm) const;
	// [gʒu̎擾
	D3DXMATRIX GetPositionMatrix(size_t frm) const;

//
// macros
private:
	/**
	 * @fn bool CMotionData::IsRange(size_t frm) const
	 * @brief l攻}N(t[)
	 * @param frm [in] Ώۃt[ԍ
	 * @retval true l
	 * @retval false lO
	 */
	bool IsRange(size_t frm) const
		{
			return frm < m_nFrames;
		}
	/**
	 * @fn bool CMotionData::IsRange(size_t frm, size_t jid) const
	 * @brief l攻}N(t[&Rx)
	 * @param frm [in] Ώۃt[ԍ
	 * @param jid [in] Ώۊ֐߃CfNX
	 * @retval true l
	 * @retval false lO
	 */
	bool IsRange(size_t frm, size_t jid) const
		{
			return frm < m_nFrames && jid < m_nJoints;
		}
	/**
	 * @fn D3DXQUATERNION& CMotionData::ROT(size_t frm, size_t jid)
	 * @brief ]f[^ANZX}N
	 * @param frm [in] t[ԍ
	 * @param jid [in] ֐߃CfNX
	 * @return ]f[^ւ̎Q
	 */
	D3DXQUATERNION& ROT(size_t frm, size_t jid)
		{
			return *(m_pRotation + m_nJoints * frm + jid);
		}
	/**
	 * @fn const D3DXQUATERNION& CMotionData::ROT(size_t frm, size_t jid) const
	 * @brief ]f[^ANZX}N(const)
	 * @param frm [in] t[ԍ
	 * @param jid [in] ֐߃CfNX
	 * @return ]f[^ւ̎Q(const)
	 */
	const D3DXQUATERNION& ROT(size_t frm, size_t jid) const
		{
			return *(m_pRotation + m_nJoints * frm + jid);
		}
	/**
	 * @fn D3DXVECTOR3& CMotionData::POS(size_t frm)
	 * @brief [gʒuf[^ANZX}N
	 * @param frm [in] t[ԍ
	 * @return ʒuf[^ւ̎Q
	 */
	D3DXVECTOR3& POS(size_t frm)
		{
			return *(m_pPosition + frm);
		}
	/**
	 * @fn const D3DXVECTOR3& CMotionData::POS(size_t frm) const
	 * @brief [gʒuf[^ANZX}N(const)
	 * @param frm [in] t[ԍ
	 * @return ʒuf[^ւ̎Q(const)
	 */
	const D3DXVECTOR3& POS(size_t frm) const
		{
			return *(m_pPosition + frm);
		}
};

#endif //MOTION_DATA_CLASS_HEADER__
