/**
 * @file MotionSignal.cpp
 * @brief NMotionSignalOԂ̎
 * @author cherub
 */
#include "stdafx.h"
#include "MotionSignal.h"
#include "MotionData.h"
#include "DistanceMatrix.h"
#include "TimewarpMatrix.h"
#include "Figure.h"
#include "FigureUtil.h"

static D3DXQUATERNION calc_differential_expmap(const D3DXQUATERNION &q0, const D3DXQUATERNION &q1);

/**
 * @fn D3DXQUATERNION calc_differential_expmap(const D3DXQUATERNION &q0, const D3DXQUATERNION &q1)
 * @brief 2̃NH[^jI̍exponential map ̌vZ
 * @param q0 [in] NH[^jI0
 * @param q1 [in] NH[^jI1
 * @return exponential map : Ln(q0^{-1} * q1)
 */
D3DXQUATERNION calc_differential_expmap(const D3DXQUATERNION &q0, const D3DXQUATERNION &q1)
{
	D3DXQUATERNION q0i;
	D3DXQuaternionInverse(&q0i, &q0);
	D3DXQUATERNION qd = q0i * q1;

	// FD3DXQuaternionLn ֐p exp map vZOɂ́Cw 𐳒lɐK
	if (qd.w < 0)
		qd *= -1.0f;
	D3DXQUATERNION expm;
	D3DXQuaternionLn(&expm, &qd);

	return expm;
}

/**
 * @fn D3DXQUATERNION interpolate_rotation(const std::vector<D3DXQUATERNION> &rots, const std::vector<float> &kernel)
 * @brief ]̕ԁDJ[l̑a1.0ɐKD
 * @param rots [in] ]
 * @param kernel [in] ԃJ[l
 * @return Ԍ
 */
D3DXQUATERNION NMotionSignal::InterpolateRotations(const std::vector<D3DXQUATERNION> &rots, const std::vector<float> &kernel)
{
	// J[la
	float ksum = 0;
	for (size_t k = 0;k < kernel.size();++k)
		ksum += kernel[k];

	D3DXQUATERNION qd, exd(0, 0, 0, 0);
	for (size_t k = 1;k < kernel.size();++k)
		exd += kernel[k] * calc_differential_expmap(rots[0], rots[k]);
	exd /= ksum;
	D3DXQuaternionExp(&qd, &exd);
	return rots[0] * qd;
}

/**
 * @fn bool NMotionSignal::Filter(CMotionData &dest, const std::vector<float> &kernel)
 * @brief [Vf[^Ŝɑ΂ꊇtB^O
 * @param dest [in/out] Ώۃ[Vf[^
 * @param kernel [in] tB^J[l
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::Filter(CMotionData &dest, const std::vector<float> &kernel)
{
	if (!dest.IsActive() || kernel.empty())
		return false;
	if (kernel.size() % 2 == 0 || dest.NumFrames() < kernel.size())
		return false;

	// J[l̑aCJ[ľvZ
	float ksum = 0;
	for (size_t k = 0;k < kernel.size();++k)
		ksum += kernel[k];
	size_t kernel_width = kernel.size() / 2;

	std::vector<D3DXQUATERNION> rots(kernel.size());
	CMotionData tmp = dest;
	for (size_t f = kernel_width;f < dest.NumFrames() - kernel_width;++f)
	{
		// [gʒuɑ΂tB^O: ʓIȏݍ݌vZ
		D3DXVECTOR3 pos(0, 0, 0);
		for (size_t k = 0;k < kernel.size();++k)
			pos += tmp.GetPosition(f + k - kernel_width) * kernel[k];
		dest.SetPosition(f, pos / ksum);

		// e֐߉]ɑ΂tB^O: CN^tB^O
		for (size_t j = 0;j < dest.NumJoints();++j)
		{
			for (size_t k = 0;k < kernel.size();++k)
				rots[k] = tmp.GetRotation(f + k - kernel_width, j);
			dest.SetRotation(f, j, InterpolateRotations(rots, kernel));
		}
	}
	return true;
}

/**
 * @fn bool NMotionSignal::Interpolate(CMotionData &dest, const CMotionData &src1, const CMotionData &src2, float ratio)
 * @brief 2̃[Vf[^̕: dest = (1.0f - ratio) * src1 + rate * src2D
 * @param dest [out] Ԃꂽ[Vf[^
 * @param src1 [in] [Vf[^1
 * @param src2 [in] [Vf[^2
 * @param ratio [in] ԌW
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::Interpolate(CMotionData &dest, const CMotionData &src1, const CMotionData &src2, float ratio)
{
	if (!src1.IsActive() || !src2.IsActive())
		return false;
	if (src1.NumJoints() != src2.NumJoints())
		return false;

	dest.Initialize(src1.NumFrames() < src2.NumFrames() ? src1.NumFrames() : src2.NumFrames(), src1.NumJoints());
	for (size_t f = 0;f < dest.NumFrames();++f)
	{
		dest.SetPosition(f, (1.0f - ratio) * src1.GetPosition(f) + ratio * src2.GetPosition(f));
		for (size_t j = 0;j < dest.NumJoints();++j)
		{
			D3DXQUATERNION q1 = src1.GetRotation(f, j);
			D3DXQUATERNION q2 = src2.GetRotation(f, j);
			D3DXQUATERNION qc;
			D3DXQuaternionSlerp(&qc, &q1, &q2, ratio);
			dest.SetRotation(f, j, qc);
		}
	}
	return true;
}

/**
 * @fn bool NMotionSignal::Interpolate(CMotionData &dest, const std::vector<CMotionData*> &srcs, const std::vector<float> &kernel)
 * @brief ̃[Vf[^̕: dest = kernel[i] * srcs[i]D
 * @param dest [out] Ԃꂽ[Vf[^
 * @param srcs [in] Tv[Vf[^
 * @param kernel [in] ԃJ[l
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::Interpolate(CMotionData &dest, const std::vector<CMotionData*> &srcs, const std::vector<float> &kernel)
{
	if (kernel.empty() || kernel.size() != srcs.size())
		return false;

	float ksum = 0;
	for (size_t k = 0;k < kernel.size();++k)
		ksum += kernel[k];

	size_t frames = srcs[0]->NumFrames(), joints = srcs[0]->NumJoints();
	for (size_t m = 1;m < srcs.size();++m)
	{
		if (joints != srcs[m]->NumJoints())
			return false;
		if (frames > srcs[m]->NumFrames())
			frames = srcs[m]->NumFrames();
	}
	dest.Initialize(frames, joints);

	std::vector<D3DXQUATERNION> rots(kernel.size());
	for (size_t f = 0;f < frames;++f)
	{
		D3DXVECTOR3 pos(0, 0, 0);
		for (size_t k = 0;k < kernel.size();++k)
			pos += kernel[k] * srcs[k]->GetPosition(f);
		dest.SetPosition(f, pos / ksum);

		for (size_t j = 0;j < joints;++j)
		{
			for (size_t k = 0;k < kernel.size();++k)
				rots[k] = srcs[k]->GetRotation(f, j);
			dest.SetRotation(f, j, InterpolateRotations(rots, kernel));
		}
	}
	return true;
}

/**
 * @fn bool NMotionSignal::Synchronize(CMotionData &dest, CVector &iitw, const CMotionData &src, const CMotionData &base, const CFigure &fig)
 * @brief 2̓̓
 * @param dest [out] baseɓsrc̎ԕϊ
 * @param iitw [out] QIԋtϊV[PX
 * @param src [in] ԕϊΏۓ
 * @param base [in] QƓ
 * @param fig [in] ƗptBMA
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::Synchronize(CMotionData &dest, CVector &iitw, const CMotionData &src, const CMotionData &base, const CFigure &fig)
{
	CFigure figs = fig, figb = fig;

	// švZ
	CDistanceMatrix dmat(src.NumFrames(), base.NumFrames());
	for (size_t i = 0;i < src.NumFrames();++i)
	{
		figs.SetPose(&src, i);
		for (size_t j = 0;j < base.NumFrames();++j)
		{
			figb.SetPose(&base, j);
			dmat(i, j) = NFigureUtil::PoseDistance(figs, figb);
		}
	}

	// registration curvěvZ
	std::vector<std::pair<int, int>> path = dmat.CalcRegistrationCurve();
	// timewarpsւ̕ϊ
	CTimewarpMatrix twm(path);
	// ԕϊ
	dest = twm.Transform(src);

	//
	// QIԋtϊV[PX
	iitw.Initialize(src.NumFrames());
	for (size_t f = 0;f < iitw.Size();++f)
		iitw[f] = static_cast<double>(f);
	iitw = twm.Transform(iitw);

	for (size_t f = 0;f < iitw.Size() - 1;++f)
		iitw[f] = iitw[f + 1] - iitw[f];
	iitw[iitw.Size() - 1] = 1.0;

	return true;
}

/**
 * @fn bool NMotionSignal::Synchronize(CMotionData &dest, CVector &iitw, const CMotionData &src, const CMotionData &base, const CFigure &fig)
 * @brief g̉^p2̓̓
 * @param dest [out] baseɓsrc̎ԕϊ
 * @param iitw [out] QIԋtϊV[PX
 * @param src [in] ԕϊΏۓ
 * @param base [in] QƓ
 * @param fig [in] ƗptBMA
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::SynchronizeLowerBody(CMotionData &dest, CVector& iitw, const CMotionData &src, const CMotionData &base, const CFigure &fig)
{
	CFigure figs = fig, figb = fig;

	// ĝ݂ΏۂƂ鋗švZ
	CDistanceMatrix dmat(src.NumFrames(), base.NumFrames());
	for (size_t i = 0;i < src.NumFrames();++i)
	{
		figs.SetPose(&src, i);
		for (size_t j = 0;j < base.NumFrames();++j)
		{
			figb.SetPose(&base, j);
			dmat(i, j) = NFigureUtil::PoseDistanceLowerBody(figs, figb);
		}
	}

	// registration curvěvZ
	std::vector<std::pair<int, int>> path = dmat.CalcRegistrationCurve();
	// timewarpsւ̕ϊ
	CTimewarpMatrix twm(path);
	// ԕϊ
	dest = twm.Transform(src);

	//
	// QIԋtϊV[PX
	iitw.Initialize(src.NumFrames());
	for (size_t f = 0;f < iitw.Size();++f)
		iitw[f] = static_cast<double>(f);
	iitw = twm.Transform(iitw);

	for (size_t f = 0;f < iitw.Size() - 1;++f)
		iitw[f] = iitw[f + 1] - iitw[f];
	iitw[iitw.Size() - 1] = 1.0;

	return true;
}

/**
 * @fn bool NMotionSignal::IncrementalTimewarp(CMotionData &dest, const CMotionData &src, const CVector &itw)
 * @brief QIԕϊ
 * @param dest [out] ʂ̓
 * @param src [in] Ώۓ
 * @param itw [in] QIԕϊV[PX
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::IncrementalTimewarp(CMotionData &dest, const CMotionData &src, const CVector &itw)
{
	double frm = 0;
	for (size_t i = 0;i < itw.Size();++i)
		frm += itw[i];
	dest.Initialize(static_cast<size_t>(frm), src.NumJoints());

	// t[̎p͂̂܂܃Rs[
	dest.SetPosition(0, src.GetPosition(0));
	for (size_t j = 0;j < src.NumJoints();j++)
		dest.SetRotation(0, j, src.GetRotation(0, j));
	
	double back = 0, pos = itw[0];
	for (size_t i = 1;i < itw.Size();++i)
	{
		size_t f0 = static_cast<size_t>(ceil(back));
		size_t f1 = static_cast<size_t>(floor(pos));
		
		for (size_t f = f0;f <= f1;++f)
		{
			float t = (static_cast<float>(f) - static_cast<float>(back)) / static_cast<float>(pos - back);
			
			D3DXVECTOR3 v1 = src.GetPosition(i - 1), v2 = src.GetPosition(i);
			dest.SetPosition(f, v1 * (1.0f - t) + t * v2);
			
			for (size_t j = 0;j < src.NumJoints();++j)
			{
				D3DXQUATERNION q, q1 = src.GetRotation(i - 1, j), q2 = src.GetRotation(i, j);
				D3DXQuaternionSlerp(&q, &q1, &q2, t);
				dest.SetRotation(f, j, q);
			}
		}
		
		if (f1 - f0 >= 0)
			back = pos;
		pos += itw[i];
	}
	return true;
}

/**
 * @fn bool NMotionSignal::Blend(CMotionData &dest, const CMotionData &src1, const CMotionData &src2, float ratio, const CFigure &fig)
 * @brief 2̓̃ufBO
 * @param dest [out] ufBO
 * @param src1 [in] Ώۓ1
 * @param src2 [in] Ώۓ2
 * @param ratio [in] uhW: dest = (1.0 - ratio) * src1 + ratio * src2
 * @param fig [in] ƗptBMA
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::Blend(CMotionData &dest, const CMotionData &src1, const CMotionData &src2, float ratio, const CFigure &fig)
{
	CMotionData tmp, blended;
	CVector iitw, fill1(src2.NumFrames(), 1.0);
	
	Synchronize(tmp, iitw, src1, src2, fig);
	Interpolate(blended, tmp, src2, ratio);

	iitw = iitw * (1.0 - ratio) + fill1 * ratio;
	IncrementalTimewarp(dest, blended, iitw);

	return true;
}

/**
 * @fn bool NMotionSignal::Blend(CMotionData &dest, const std::vector<CMotionData*> &srcs, const std::vector<float> &kernel, const CFigure &fig)
 * @brief ̓̃ufBO
 * @param dest [out] ufBO
 * @param srcs [in] ΏۓQ
 * @param kernel [in] ԃJ[l
 * @param fig [in] ƗptBMA
 * @retval true 
 * @retval false s
 */
bool NMotionSignal::Blend(CMotionData &dest, const std::vector<CMotionData*> &srcs, const std::vector<float> &kernel, const CFigure &fig)
{
	CMotionData blended;
	std::vector<CMotionData*> mtmp;
	CVector vtmp, iitw(srcs.front()->NumFrames(), kernel.front());

	mtmp.push_back(srcs.front());
	for (size_t i = 1;i < srcs.size();++i)
	{
		mtmp.push_back(new CMotionData);
		Synchronize(*mtmp.back(), vtmp, *srcs[i], *srcs[0], fig);
		iitw += vtmp * kernel[i];
	}

	Interpolate(blended, mtmp, kernel);
	IncrementalTimewarp(dest, blended, iitw);

	while (mtmp.size() > 1)
	{
		delete mtmp.back();
		mtmp.pop_back();
	}
	return true;
}
