// SceneView.cpp : CSceneView NX̎
//

#include "stdafx.h"
#include "MotionSplicing.h"

#include "MainDoc.h"
#include "SceneView.h"

#include "FigureUtil.h"
#include "MotionFile.h"
#include "MotionData.h"
#include "MotionSignal.h"
#include "DistanceMatrix.h"
#include "Statistics.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


// CSceneView

IMPLEMENT_DYNCREATE(CSceneView, CView)

BEGIN_MESSAGE_MAP(CSceneView, CView)
	ON_MESSAGE(WM_USER_RENDER_SCENE, OnRenderScene)
	ON_WM_DESTROY()
	ON_WM_PAINT()
	ON_WM_SIZE()
	ON_WM_LBUTTONDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_MBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_WM_TIMER()
	ON_COMMAND(ID_FILE_OPEN, &CSceneView::OnFileOpen)
	ON_COMMAND(ID_FILE_PLAY, &CSceneView::OnFilePlay)
	ON_UPDATE_COMMAND_UI(ID_FILE_PLAY, &CSceneView::OnUpdateFilePlay)
	ON_COMMAND(ID_FILE_SAVE_AS, &CSceneView::OnFileSaveAs)
	ON_COMMAND(ID_MOTION_TEMPORAL_ALIGN, &CSceneView::OnMotionTemporalAlign)
	ON_COMMAND(ID_MOTION_SPATIAL_ALIGN, &CSceneView::OnMotionSpatialAlign)
	ON_COMMAND(ID_MOTION_SPLICING, &CSceneView::OnMotionSplicing)
	ON_COMMAND(ID_MOTION_TRANSFER, &CSceneView::OnMotionTransfer)
END_MESSAGE_MAP()

// CSceneView RXgNV/fXgNV

CSceneView::CSceneView()
{
	m_pD3D = NULL;
	m_pDeviceD3D = NULL;
	m_pEffect = NULL;

	m_hWorld = NULL;
	m_hWorldIT = NULL;
	m_hWorldViewProj = NULL;
	m_hEyePos = NULL;

	m_pvView[0].x = 5.0f;
	m_pvView[0].y = 0.0f;
	m_pvView[0].z = 0.0f;

	m_pvView[1].x = 0.0f;
	m_pvView[1].y = 0.0f;
	m_pvView[1].z = 0.0f;

	m_pvView[2].x = 0.0f;
	m_pvView[2].y = 1.0f;
	m_pvView[2].z = 0.0f;

	for (int i = 0;i < 3;i++)
		D3DXMatrixIdentity(&m_pmViewTrans[i]);

	m_pMeshSphere = NULL;
	for (size_t i = 0;i < 3;++i)
	{
		m_pFigure[i] = NULL;
		m_pMotion[i] = NULL;
	}
	m_frame = 0;
	m_bPlay = false;
}

CSceneView::~CSceneView()
{
}

BOOL CSceneView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: ̈ʒu CREATESTRUCT cs C Window NX܂̓X^C
	//  CĂB

	return CView::PreCreateWindow(cs);
}

// CSceneView `

void CSceneView::OnDraw(CDC* /*pDC*/)
{
	CMainDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	if (!pDoc)
		return;

	// TODO: ̏ꏊɃlCeBu f[^p̕`R[hǉ܂B
}


// CSceneView ff

#ifdef _DEBUG
void CSceneView::AssertValid() const
{
	CView::AssertValid();
}

void CSceneView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CMainDoc* CSceneView::GetDocument() const // fobOȊÕo[W̓CCłB
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CMainDoc)));
	return (CMainDoc*)m_pDocument;
}
#endif //_DEBUG


// CSceneView bZ[W nh

void CSceneView::OnInitialUpdate()
{
	CView::OnInitialUpdate();

	HRESULT hr;

	if (hr = InitDirectGraphics(), FAILED(hr))
	{
		MessageBox(DXGetErrorString9(hr));
		exit(hr);
	}

	if (hr = InitRenderState(), FAILED(hr))
	{
		MessageBox(DXGetErrorString9(hr));
		exit(hr);
	}

	if (hr = InitHLSL(), FAILED(hr))
	{
		MessageBox(DXGetErrorString9(hr));
		exit(hr);
	}

	D3DXCreateSphere(m_pDeviceD3D, 0.5f, 10, 10, &m_pMeshSphere, NULL);

	SetTimer(0, 30, NULL);
}

void CSceneView::OnDestroy()
{
	CView::OnDestroy();

	for (size_t i = 0;i < 3;++i)
	{
		if (m_pFigure[i] != NULL)
			delete m_pFigure[i];
		if (m_pMotion[i] != NULL)
			delete m_pMotion[i];
	}

	if (m_pMeshSphere != NULL)
		m_pMeshSphere->Release();
	if (m_pEffect != NULL)
		m_pEffect->Release();
	if (m_pDeviceD3D != NULL)
		m_pDeviceD3D->Release();
	if (m_pD3D != NULL)
		m_pD3D->Release();
}

void CSceneView::OnPaint()
{
	CPaintDC dc(this);

	SendMessage(WM_USER_RENDER_SCENE);
}

void CSceneView::OnSize(UINT nType, int cx, int cy)
{
	CView::OnSize(nType, cx, cy);

	if (m_pD3D == NULL)
		return;

	D3DDISPLAYMODE d3dDM;
	HRESULT hr;

	if (hr = m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3dDM), FAILED(hr))
		return;

	float fAspect = ((float)cx / (float)cy) * ((d3dDM.Height * 4.0f) / (d3dDM.Width * 3.0f));
	D3DXMatrixPerspectiveFovLH(&m_mProjection, D3DX_PI / 4.0f, fAspect, 1.0f, 100.0f);
}

HRESULT CSceneView::InitDirectGraphics(void)
{
	HRESULT hr;
	D3DDISPLAYMODE d3dDM;
	D3DPRESENT_PARAMETERS d3dPP;

	if (m_pD3D = Direct3DCreate9(D3D_SDK_VERSION), m_pD3D == NULL)
		return E_FAIL;

	if (hr = m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3dDM), FAILED(hr))
		return hr;

	ZeroMemory(&d3dPP, sizeof(D3DPRESENT_PARAMETERS));
	d3dPP.Windowed = TRUE;
	d3dPP.BackBufferWidth = 640;
	d3dPP.BackBufferHeight = 480;
	d3dPP.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dPP.BackBufferFormat = d3dDM.Format;
	d3dPP.EnableAutoDepthStencil = TRUE;
	d3dPP.AutoDepthStencilFormat = D3DFMT_D16;
	d3dPP.BackBufferCount = 1;
	d3dPP.Flags |= D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;

    if (hr = m_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, GetSafeHwnd(), D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dPP, &m_pDeviceD3D), FAILED(hr))
	{
		if (hr = m_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, GetSafeHwnd(), D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dPP, &m_pDeviceD3D), FAILED(hr))
		{
			if (hr = m_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, GetSafeHwnd(), D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dPP, &m_pDeviceD3D), FAILED(hr))
			{
				return hr;
			}
		}
	}
	return S_OK;
}

HRESULT CSceneView::InitRenderState(void)
{
	HRESULT hr;
	D3DDISPLAYMODE d3dDM;
	D3DXVECTOR3 pvTmp[3];

	m_pDeviceD3D->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	m_pDeviceD3D->SetRenderState(D3DRS_DITHERENABLE, FALSE);
	m_pDeviceD3D->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, FALSE);
	m_pDeviceD3D->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

	D3DVIEWPORT9 vp;
	if(hr = m_pDeviceD3D->GetViewport(&vp), FAILED(hr))
		return hr;

	if (hr = m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3dDM), FAILED(hr))
		return hr;

	float fAspect = ((float)vp.Width / (float)vp.Height) * ((d3dDM.Height * 4.0f) / (d3dDM.Width * 3.0f));
	D3DXMatrixPerspectiveFovLH(&m_mProjection, D3DX_PI / 4.0f, fAspect, 1.0f, 100.0f);

	for (int i = 0;i < 3;i++)
		D3DXVec3TransformCoord(&pvTmp[i], &m_pvView[i], &m_pmViewTrans[i]);
	D3DXMatrixLookAtLH(&m_mView, &pvTmp[0], &pvTmp[1], &pvTmp[2]);

	return hr;
}

HRESULT CSceneView::InitHLSL(void)
{
	HRESULT	hr;

	if (hr = D3DXCreateEffectFromFile(m_pDeviceD3D, "Viewer.fx",
					NULL, NULL, 0, NULL, &m_pEffect, NULL), FAILED(hr))
		return hr;
	
	m_hTechPhong = m_pEffect->GetTechniqueByName("PhongShader");
	if (m_hTechPhong == NULL)
		return E_FAIL;

	if (m_hWorld = m_pEffect->GetParameterByName(NULL, "g_mWorld"), m_hWorld == NULL)
		return E_FAIL;
	if (m_hWorldIT = m_pEffect->GetParameterByName(NULL, "g_mWorldIT"), m_hWorldIT == NULL)
		return E_FAIL;
	if (m_hWorldViewProj = m_pEffect->GetParameterByName(NULL, "g_mWorldViewProj"), m_hWorldViewProj == NULL)
		return E_FAIL;
	if (m_hEyePos = m_pEffect->GetParameterByName(NULL, "g_vEyePos"), m_hEyePos == NULL)
		return E_FAIL;

	D3DXVECTOR4 vTmp(m_pvView->x, m_pvView->y, m_pvView->z, 1.0f);
	m_pEffect->SetVector(m_hEyePos, &vTmp);

	return hr;
}

void CSceneView::OnLButtonDown(UINT nFlags, CPoint point)
{
	m_pntDown = point;
	CView::OnLButtonDown(nFlags, point);
}

void CSceneView::OnRButtonDown(UINT nFlags, CPoint point)
{
	m_pntDown = point;
	CView::OnRButtonDown(nFlags, point);
}

void CSceneView::OnMButtonDown(UINT nFlags, CPoint point)
{
	m_pntDown = point;
	CView::OnMButtonDown(nFlags, point);
}

void CSceneView::OnMouseMove(UINT nFlags, CPoint point)
{
	D3DXMATRIX mTrans, mTmp;
	D3DXVECTOR3 pvTmp[3];
	D3DXQUATERNION qt;

	int dx = point.x - m_pntDown.x;
	int dy = point.y - m_pntDown.y;

	switch (nFlags)
	{
	case MK_SHIFT | MK_LBUTTON:
		D3DXMatrixRotationY(&mTrans, (float)dx / 400.0f);
		mTrans *= *D3DXMatrixRotationZ(&mTmp, (float)dy / 400.0f);
		m_pmViewTrans[0] = mTrans * m_pmViewTrans[0];
		m_pmViewTrans[1] = mTrans * m_pmViewTrans[1];
		m_pmViewTrans[2] = mTrans * m_pmViewTrans[2];
		break;
	case MK_SHIFT | MK_RBUTTON:
		D3DXMatrixTranslation(&mTrans, 0.0f, dy / 40.0f, -dx / 40.0f);
		m_pmViewTrans[0] = mTrans * m_pmViewTrans[0];
		m_pmViewTrans[1] = mTrans * m_pmViewTrans[1];
		break;
	case MK_SHIFT | MK_MBUTTON:
		m_pvView[0].x += -dy / 100.0f;
		if (m_pvView[0].x < 0.1f)
			m_pvView[0].x = 0.1f;
		break;
	default:
		return;
	}

	for (int i = 0;i < 3;i++)
		D3DXVec3TransformCoord(&pvTmp[i], &m_pvView[i], &m_pmViewTrans[i]);
	D3DXMatrixLookAtLH(&m_mView, &pvTmp[0], &pvTmp[1], &pvTmp[2]);

	D3DXVECTOR4 vTmp(pvTmp->x, pvTmp->y, pvTmp->z, 1.0f);
	m_pEffect->SetVector(m_hEyePos, &vTmp);

	m_pntDown = point;

	SendMessage(WM_USER_RENDER_SCENE);

	CView::OnMouseMove(nFlags, point);
}

BOOL CSceneView::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
	m_pvView[0].x += static_cast<float>(zDelta) / 60.0f;
	if (m_pvView[0].x < 0.1f)
		m_pvView[0].x = 0.1f;

	D3DXVECTOR3 pvTmp[3];
	for (int i = 0;i < 3;i++)
		D3DXVec3TransformCoord(&pvTmp[i], &m_pvView[i], &m_pmViewTrans[i]);
	D3DXMatrixLookAtLH(&m_mView, &pvTmp[0], &pvTmp[1], &pvTmp[2]);

	SendMessage(WM_USER_RENDER_SCENE);

	return CView::OnMouseWheel(nFlags, zDelta, pt);
}

LRESULT CSceneView::OnRenderScene(WPARAM wPara, LPARAM lPara)
{
	HRESULT hr;
	m_pDeviceD3D->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0f, 0);

	if (hr = m_pDeviceD3D->BeginScene(), FAILED(hr))
		return -1;

	m_pEffect->SetTechnique("PhongShader");
	m_pEffect->SetVector("g_vAmbiColor", &D3DXVECTOR4(0.8f, 0.8f, 0.5f, 1.0f));
	m_pEffect->SetVector("g_vSurfColor", &D3DXVECTOR4(0.8f, 0.8f, 0.5f, 1.0f));
	DrawSkeleton(0);

	m_pEffect->SetVector("g_vAmbiColor", &D3DXVECTOR4(0.5f, 0.8f, 0.8f, 1.0f));
	m_pEffect->SetVector("g_vSurfColor", &D3DXVECTOR4(0.5f, 0.8f, 0.8f, 1.0f));
	DrawSkeleton(1);

	m_pEffect->SetVector("g_vAmbiColor", &D3DXVECTOR4(0.8f, 0.5f, 0.8f, 1.0f));
	m_pEffect->SetVector("g_vSurfColor", &D3DXVECTOR4(0.8f, 0.5f, 0.8f, 1.0f));
	DrawSkeleton(2);

	if (hr = m_pDeviceD3D->EndScene(), FAILED(hr))
		return -1;

	hr = m_pDeviceD3D->Present(NULL, NULL, NULL, NULL);

	return 0;
}

HRESULT CSceneView::DrawMeshSub(const LPD3DXMESH pMesh, const D3DXMATRIX &mWorld) const
{
	HRESULT hr;
	unsigned int nNumPass;
	D3DXMATRIX mTmp;

	m_pEffect->SetMatrix(m_hWorld, &mWorld);
	D3DXMatrixInverse(&mTmp, NULL, &mWorld);
	D3DXMatrixTranspose(&mTmp, &mTmp);
	m_pEffect->SetMatrix(m_hWorldIT, &mTmp);

	mTmp = mWorld * m_mView * m_mProjection;
	m_pEffect->SetMatrix(m_hWorldViewProj, &mTmp);

	if (hr = m_pEffect->Begin(&nNumPass, 0), FAILED(hr))
		return hr;

	m_pEffect->BeginPass(0);
	pMesh->DrawSubset(0);
	m_pEffect->EndPass();

	m_pEffect->End();

	return S_OK;
}

void CSceneView::DrawSkeletonSub(const CJoint *pJoint, LPD3DXMATRIXSTACK &pMatStack)
{
	if (pJoint == NULL)
		return;

	pMatStack->Push();
		pMatStack->MultMatrixLocal(&pJoint->GetInitialTransform());
		DrawMeshSub(m_pMeshSphere, *pMatStack->GetTop());
	pMatStack->Pop();

	pMatStack->Push();
		pMatStack->MultMatrixLocal(&pJoint->GetTransformMatrix());
		for (size_t j = 0;j < pJoint->NumChildren();j++)
			DrawSkeletonSub(pJoint->GetChild(j), pMatStack);
	pMatStack->Pop();
}

void CSceneView::DrawSkeleton(size_t id)
{
	if (m_pFigure[id] == NULL)
		return;

	HRESULT hr;
	LPD3DXMATRIXSTACK pMatStack;
	const CJoint *pRoot = m_pFigure[id]->GetRoot();;

	if (hr = D3DXCreateMatrixStack(0, &pMatStack), FAILED(hr))
		return;

	pMatStack->LoadIdentity();
	pMatStack->Scale(0.01f, 0.01f, 0.01f);

	pMatStack->MultMatrixLocal(&pRoot->GetTransformMatrix());
	for (size_t j = 0;j < pRoot->NumChildren();j++)
		DrawSkeletonSub(pRoot->GetChild(j), pMatStack);

	pMatStack->Release();
}

void CSceneView::OnTimer(UINT_PTR nIDEvent)
{
	size_t cnt = 0;
	bool flag = false;
	for (size_t i = 0;m_bPlay && i < 3;++i)
	{
		if (m_pFigure[i] == NULL || m_pMotion[i] == NULL)
			continue;

		if (m_frame < m_pMotion[i]->NumFrames())
		{
			m_pFigure[i]->SetPose(m_pMotion[i], m_frame);
			flag = true;
		}
		else
		{
			m_pFigure[i]->SetPose(m_pMotion[i], m_pMotion[i]->NumFrames() - 1);
		}
	}
	m_frame = flag ? m_frame + 2 : 0;

	SendMessage(WM_USER_RENDER_SCENE);

	CView::OnTimer(nIDEvent);
}

void CSceneView::OnFileOpen()
{
	CFileDialog dlg(TRUE, "bvh", NULL, 0, "BVH Motion File (*.bvh)|*.bvh||", this);

	for (size_t i = 0;i < 2;++i)
	{
		if (dlg.DoModal() != IDOK)
			continue;

		if (m_pMotion[i] != NULL)
			delete m_pMotion[i];
		m_pMotion[i] = new CMotionData;

		if (m_pFigure[i] != NULL)
			delete m_pFigure[i];
		m_pFigure[i] = new CFigure;

		CWaitCursor cur;
		if (NMotionFile::LoadBvhFile(dlg.GetPathName(), m_pFigure[i], m_pMotion[i]))
		{
			m_frame = 0;
			m_pFigure[i]->SetPose(m_pMotion[i], 0);
			continue;
		}

		delete m_pMotion[i];
		m_pMotion[i] = NULL;
		delete m_pFigure[i];
		m_pFigure[i] = NULL;
	}
}

void CSceneView::OnFileSaveAs()
{
	CFileDialog dlg(TRUE, "bvh", NULL, 0, "BVH Motion File (*.bvh)|*.bvh||", this);

	if (m_pFigure[2] == NULL || m_pMotion[2] == NULL || dlg.DoModal() != IDOK)
		return;

	NMotionFile::StoreBvhFile(dlg.GetPathName(), m_pFigure[2], m_pMotion[2]);
}

void CSceneView::OnFilePlay()
{
	m_bPlay = !m_bPlay;
}

void CSceneView::OnUpdateFilePlay(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(m_bPlay);
}

/**
 * @fn void CSceneView::OnMotionTemporalAlign()
 * @brief g̉^f[^݂̂p A ̓ B ւ̓ 
 */
void CSceneView::OnMotionTemporalAlign()
{
	CWaitCursor cur;
	
	CMotionData dst;
	CVector itw;
	//  A ̎ԕϊiԋtϊ͕svȂ̂łŔjj
	NMotionSignal::SynchronizeLowerBody(dst, itw, *m_pMotion[1], *m_pMotion[0], *m_pFigure[0]);
	*m_pMotion[1] = dst;

	//  B 𓮍 C ɃRs[
	if (m_pFigure[2] != NULL)
		delete m_pFigure[2];
	m_pFigure[2] = new CFigure(*m_pFigure[0]);

	if (m_pMotion[2] != NULL)
		delete m_pMotion[2];
	m_pMotion[2] = new CMotionData(*m_pMotion[0]);

	for (size_t i = 0;i < 3;++i)
		m_pFigure[i]->SetPose(m_pMotion[i], m_frame);
	SendMessage(WM_USER_RENDER_SCENE);
}

/**
 * @fn void CSceneView::OnMotionSpatialAlign()
 * @brief eԃt[ƂɓƗ MU  pelvis ]␳
 */
void CSceneView::OnMotionSpatialAlign()
{
	// pelvism[ȟ
	const size_t tid = m_pFigure[0]->FindNode("vl5");
	for (size_t frm = 0;frm < m_pMotion[0]->NumFrames();++frm)
	{
		m_pFigure[0]->SetPose(m_pMotion[0], frm);
		m_pFigure[1]->SetPose(m_pMotion[1], frm);

		// 㔼g_Q̎Zo
		std::vector<D3DXVECTOR3> pc0 = m_pFigure[0]->GetPointCloudUpperBody();
		std::vector<D3DXVECTOR3> pc1 = m_pFigure[1]->GetPointCloudUpperBody();

		// pelvis ̃[JWn[hWn_ɈvgXtH[
		D3DXMATRIX lp0, lp1;
		D3DXMatrixInverse(&lp0, NULL, &m_pFigure[0]->GetWorldCoordinate(tid));
		D3DXMatrixInverse(&lp1, NULL, &m_pFigure[1]->GetWorldCoordinate(tid));
		for (size_t i = 0;i < pc0.size();++i)
		{
			D3DXVECTOR3 vt;
			pc0[i] = *D3DXVec3TransformCoord(&vt, &pc0[i], &lp0);
			pc1[i] = *D3DXVec3TransformCoord(&vt, &pc1[i], &lp1);
		}

		// rot(pc1)  pc0 Ȃ]gXtH[ qm
		D3DXQUATERNION qm = NFigureUtil::CalcRotationBetweenPointClouds(pc0, pc1);
		// qm ɂ AU  pelvis ]␳ CU  pelvis ɃRs[
		D3DXQUATERNION qc = m_pMotion[1]->GetRotation(frm, tid) * qm;
		m_pMotion[2]->SetRotation(frm, tid, qc);
	}

	for (size_t i = 0;i < 3;++i)
		m_pFigure[i]->SetPose(m_pMotion[i], m_frame);
	SendMessage(WM_USER_RENDER_SCENE);
}

/**
 * @fn void CSceneView::OnMotionSplicing()
 * @brief pelvis ȊO AU 㔼g̓f[^ CU ɃRs[
 */
void CSceneView::OnMotionSplicing()
{
	// g̎n[m[h
	const size_t eid = m_pFigure[1]->FindNode("sacroiliac");
	const size_t tid = m_pFigure[1]->FindNode("vl5");

	for (size_t frm = 0;frm < m_pMotion[1]->NumFrames();++frm)
	for (size_t jid = 1;jid < eid;++jid)
	{
		if (jid != tid)
		{
			D3DXQUATERNION qc = m_pMotion[1]->GetRotation(frm, jid);
			m_pMotion[2]->SetRotation(frm, jid, qc);
		}
	}

	for (size_t i = 0;i < 3;++i)
		m_pFigure[i]->SetPose(m_pMotion[i], m_frame);
	SendMessage(WM_USER_RENDER_SCENE);
}

/**
 * @fn D3DXMATRIX vector_aligning_rotation(const D3DXVECTOR3 &src, const D3DXVECTOR3 &target)
 * @brief rot(base) == target
 */
#if 1 // ver
D3DXMATRIX vector_aligning_rotation(const D3DXVECTOR3 &src, const D3DXVECTOR3 &target)
{
	D3DXVECTOR3 src2, target2;
	D3DXMATRIX yms, ymt, tmp;

	D3DXMatrixRotationY(&yms, atan2f(src.z, src.x));
	D3DXVec3TransformCoord(&src2, &src, &yms);
	D3DXMatrixRotationZ(&tmp, -atan2f(src2.y, src2.x));
	yms *= tmp;

	D3DXMatrixRotationY(&ymt, atan2f(target.z, target.x));
	D3DXVec3TransformCoord(&target2, &target, &ymt);
	D3DXMatrixRotationZ(&tmp, -atan2f(target2.y, src2.x));
	ymt *= tmp;

	return yms * *D3DXMatrixInverse(&tmp, NULL, &ymt);
}
#else //s ver.
D3DXMATRIX vector_aligning_rotation(const D3DXVECTOR3 &src, const D3DXVECTOR3 &target)
{
	D3DXVECTOR3 av, nav;
	D3DXVec3Cross(&av, &src, &target);
	D3DXVec3Normalize(&nav, &av);

	D3DXQUATERNION aq;
	float angle = acos(D3DXVec3Dot(&src, &target));
	if (fabsf(angle) < 1.0e-6f)
		D3DXQuaternionIdentity(&aq);
	else
		D3DXQuaternionRotationAxis(&aq, &nav, angle);

	D3DXMATRIX am;
	return *D3DXMatrixRotationQuaternion(&am, &aq);
}
#endif

void CSceneView::OnMotionTransfer()
{
	const size_t tid = m_pFigure[1]->FindNode("vl5");
	const float fnum = static_cast<float>(m_pMotion[1]->NumFrames());

	std::vector<D3DXVECTOR3> pc1, pc2;
	for (size_t frm = 0;frm < m_pMotion[1]->NumFrames();++frm)
	{
		m_pFigure[1]->SetPose(m_pMotion[1], frm);
		m_pFigure[2]->SetPose(m_pMotion[2], frm);

		// [gʒu
		D3DXVECTOR3 rp1 = m_pFigure[1]->GetRootPosition(), rp2 = m_pFigure[2]->GetRootPosition();
		// [g
		D3DXQUATERNION rq1 = m_pFigure[1]->GetRootOrientation(), rq2 = m_pFigure[2]->GetRootOrientation();
		D3DXMATRIX rm1, rm2;
		D3DXMatrixRotationQuaternion(&rm1, &rq1);
		D3DXMatrixRotationQuaternion(&rm2, &rq2);

		// [hWnɂ鍘xNg̎Zo
		D3DXVECTOR3 vx(1, 0, 0), hv1, hv2;
		D3DXVec3TransformCoord(&hv1, &vx, &rm1);
		D3DXVec3TransformCoord(&hv2, &vx, &rm2);

		// xNg[h+XɈv邽߂̉]gXtH[s
		D3DXMATRIX am1 = vector_aligning_rotation(hv1, vx), am2 = vector_aligning_rotation(hv2, vx);

		std::vector<D3DXVECTOR3> tpc1, tpc2;
		tpc1 = m_pFigure[1]->GetPointCloudUpperBody();
		tpc2 = m_pFigure[2]->GetPointCloudUpperBody();

		if (pc1.empty())
		{
			pc1.resize(tpc1.size(), D3DXVECTOR3(0, 0, 0));
			pc2.resize(tpc2.size(), D3DXVECTOR3(0, 0, 0));
		}

		D3DXVECTOR3 vt, va;	
		for (size_t i = 0;i < tpc1.size();++i)
		{
			vt = tpc1[i] - rp1;
			pc1[i] += *D3DXVec3TransformCoord(&va, &vt, &am1);

			vt = tpc2[i] - rp2;
			pc2[i] += *D3DXVec3TransformCoord(&va, &vt, &am2);
		}
	}
	for (size_t i = 0;i < pc1.size();++i)
	{
		pc1[i] /= fnum;
		pc2[i] /= fnum;
	}

	// CU ̏㔼g_Q AU ̓_Qɍv悤 CU  pelvis ]␳
	D3DXQUATERNION qt = NFigureUtil::CalcRotationBetweenPointClouds(pc1, pc2);
	for (size_t frm = 0;frm < m_pMotion[2]->NumFrames();++frm)
	{
		D3DXQUATERNION qc = m_pMotion[2]->GetRotation(frm, tid) * qt;
		m_pMotion[2]->SetRotation(frm, tid, qc);
	}

	for (size_t i = 0;i < 3;++i)
		m_pFigure[i]->SetPose(m_pMotion[i], m_frame);
	SendMessage(WM_USER_RENDER_SCENE);
}
