/**
 * @file Statistics.cpp
 * @brief v͊֐QOԎ
 * @author cherub
 */
#include "stdafx.h"
#include "Statistics.h"
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>

/**
 * @fn bool NStatistics::EVD(CMatrix &evecs, CVector &evals, const CMatrix &data)
 * @brief ŗLl
 * @param evecs [out] ŗLlxNg
 * @param evals [out] ŗLl
 * @param data [in] ͑Ώۃf[^siΏ̍sj
 * @retval true ͐
 * @retval false ͎s
 */
bool NStatistics::EVD(CMatrix &evecs, CVector &evals, const CMatrix &data)
{
	evecs.Initialize(data.Row(), data.Col());
	evals.Initialize(data.Row());
	CMatrix tmp = data;

	// GSLŗLl
	gsl_eigen_symmv_workspace *w = gsl_eigen_symmv_alloc(data.Row());
	int bsucc = gsl_eigen_symmv(tmp.m_pmMatrix, evals.m_pvVector, evecs.m_pmMatrix, w);
	gsl_eigen_symmv_free(w);

	if (bsucc == GSL_SUCCESS)
	{
		// ŗLl̐Βlɂč~\[g
		gsl_eigen_symmv_sort(evals.m_pvVector, evecs.m_pmMatrix, GSL_EIGEN_SORT_ABS_DESC);
		return true;
	}
	evecs.Destroy();
	evals.Destroy();
	return false;
}

/**
 * @fn bool NStatistics::SVD(CMatrix &u, CVector &s, CMatrix &v, const CMatrix &data)
 * @brief GRm~[TCYْ̓lD data = USV^T
 * @param u [out] ٍsiMxNj
 * @param s [out] ْliNj
 * @param v [out] EٍsiNxNj
 * @param data [in] ͑Ώۃf[^siTvM x ϗʐNj
 * @retval true ͐
 * @retval false ͎s
 */
bool NStatistics::SVD(CMatrix &u, CVector &s, CMatrix &v, const CMatrix &data)
{
	if (data.Row() < data.Col())
		u = data * data.Transpose();
	else if (data.Row() > data.Col())
		u = data.Transpose() * data;
	else
		u = data;

	v.Initialize(u.Col(), u.Col());
	s.Initialize(u.Col());

	gsl_vector *vW = gsl_vector_alloc(u.Col());
	int error_code = gsl_linalg_SV_decomp(u.m_pmMatrix, v.m_pmMatrix, s.m_pvVector, vW);
	gsl_vector_free(vW);

	if (error_code != GSL_SUCCESS)
	{
		u.Destroy();
		s.Destroy();
		v.Destroy();
		return false;
	}

	CMatrix is(s.Size(), s.Size(), 0);
	for (size_t i = 0;i < s.Size();++i)
	{
		s[i] = sqrt(s[i]);
		is(i, i) = s[i] > 0 ? 1.0 / s[i] : 0;
	}

	if (data.Row() < data.Col())
		v = data.Transpose() * u * is;
	else if (data.Row() > data.Col())
		u = data * v * is;

	return true;
}

/**
 * @fn bool NStatistics::PCA(CMatrix &scores, CMatrix &rot, CVector &cont, CVector &mean, const CMatrix &data)
 * @brief f[^s̒SEWʂ听
 * @param scores [out] 听XRAiMxNj
 * @param rot [out] ϊsiNxNj
 * @param cont [out] ^iMj
 * @param mean [out] ϒlxNgiNj
 * @param data [int] ͑Ώۃf[^siTvM x ϗʐNj
 * @retval true 
 * @retval false 
 */
bool NStatistics::PCA(CMatrix &scores, CMatrix &rot, CVector &cont, CVector &mean, const CMatrix &data)
{
	// S
	CMatrix nm(data.Row(), data.Row(), -1.0 / data.Row());
	for (size_t i = 0; i < data.Row(); ++i)
		nm(i, i) += 1.0;
	CMatrix centered = nm * data;

	// ϒlxNg
	mean.Initialize(data.Row(), 1.0 / data.Row());
	mean = mean * data;

	// W
	CMatrix scaled = centered;
	for (size_t c = 0;c < scaled.Col();++c)
	{
		CVector v = scaled.ColVector(c);
		if (v.Norm() > 0)
			v *= sqrt(v.Size() - 1.0) / v.Norm();
		for (size_t r = 0;r < scaled.Row();++r)
			scaled(r, c) = v[r];
	}

	// Wꂽsْ̓l
	CMatrix tmp;
	if (!NStatistics::SVD(tmp, cont, rot, scaled))
	{
		scores.Destroy();
		rot.Destroy();
		cont.Destroy();
		mean.Destroy();
		return false;
	}
	cont *= 1.0 / cont.Size();

	// 听XRA̎Zo
	scores = centered * rot;
	return true;
}
