/**
 * @file TimewarpMatrix.cpp
 * @brief sNX
 * @author cherub
 */
#include "stdafx.h"
#include "TimewarpMatrix.h"

/**
 * @fn CTimewarpMatrix::CTimewarpMatrix(void)
 * @brief ftHgRXgN^
 */
CTimewarpMatrix::CTimewarpMatrix(void)
: CMatrix()
{
}

/**
 * @fn CTimewarpMatrix::CTimewarpMatrix(const CTimewarpMatrix &src)
 * @brief Rs[RXgN^
 * @param src [in] Rs[NXCX^X
 */
CTimewarpMatrix::CTimewarpMatrix(const CTimewarpMatrix &src)
: CMatrix(src)
{
}

/**
 * @fn CTimewarpMatrix::CTimewarpMatrix(const std::vector<std::pair<int, int>> &path)
 * @brief DPŏoH񂩂̃RXgN^
 * @param path [in] DPŏoH
 */
CTimewarpMatrix::CTimewarpMatrix(const std::vector<std::pair<int, int>> &path)
{
	CTimewarpMatrix::Initialize(path);
}

/**
 * @fn CTimewarpMatrix::~CTimewarpMatrix(void)
 * @brief fXgN^
 */
CTimewarpMatrix::~CTimewarpMatrix(void)
{
}

/**
 * @fn CTimewarpMatrix& CTimewarpMatrix::operator =(const CTimewarpMatrix &src)
 * @brief Zq
 * @param src [in] NXCX^X
 * @return gւ̎Q
 */
CTimewarpMatrix& CTimewarpMatrix::operator =(const CTimewarpMatrix &src)
{
	CMatrix::operator =(src);
	return *this;
}

/**
 * @fn bool CTimewarpMatrix::Initialize(const std::vector<std::pair<int, int>> &path)
 * @brief DPŏoH񂩂̎ԕϊs񏉊
 * @param path [in] DPŏoH
 * @retval true 
 * @retval false s
 */
bool CTimewarpMatrix::Initialize(const std::vector<std::pair<int, int>> &path)
{
	CMatrix::Initialize(path.back().first + 1, path.back().second + 1, 0.0);
	size_t radj =1;
	size_t cadj =1;

	//
	// DPŏoHɉĎԕϊs𐶐
	for (std::vector<std::pair<int, int>>::const_iterator it = path.begin();it != path.end();++it)
	{
		if (it != path.end() - 1 && it->first == (it + 1)->first)
		{
			radj++;
			cadj = 1;
		}
		else if (it != path.end() - 1 && it->second == (it + 1)->second)
		{
			radj = 1;
			cadj++;
		}
		else
		{
			if (radj == 1 && cadj == 1)
			{
				Data(it->first, it->second) = 1.0;
			}
			else if (cadj > 1)
			{
				for (size_t r = 0;r < cadj;r++)
					Data(it->first - r, it->second) = 1.0 / static_cast<double>(cadj);
			}
			else if (it == path.end() - 1)
			{
				for (size_t c = 1;c <= radj;c++)
					Data(it->first, it->second - radj + c) = 1.0;
			}
			else
			{
				for (size_t c = 1;c <= radj;c++)
				{
					Data(it->first + 1, it->second - radj + c) = static_cast<double>(c - 1) / static_cast<double>(radj);
					Data(it->first, it->second - radj + c) = 1.0 - Data(it->first + 1, it->second - radj + c);
				}
			}

			radj = 1;
			cadj = 1;
		}
	}

	return true;
}

/**
 * @fn CVector CTimewarpMatrix::Transform(const CVector &vSequence) const
 * @brief Ԍoߏ̎ԕϊ
 * @param vSequence [in] Ԍoߏ
 * @return ԕϊ̎Ԍoߏ
 */
CVector CTimewarpMatrix::Transform(const CVector &vSequence) const
{
	CVector vTrans;

	if (!IsActive() || !vSequence.IsActive())
		return vTrans;

	return vSequence * *this;
}

/**
 * @fn CMotionData CTimewarpMatrix::Transform(const CMotionData &src) const
 * @brief [Vf[^̎ԕϊ
 * @param src [in] [Vf[^
 * @return ԕϊ̃[Vf[^
 */
CMotionData CTimewarpMatrix::Transform(const CMotionData &src) const
{
	CMotionData res(Col(), src.NumJoints());

	CVector seq(Row());
	for (size_t i = 0;i < Row();++i)
		seq[i] = static_cast<double>(i);
	seq = Transform(seq);

	for (size_t f = 0;f < Col();++f)
	{
		size_t f1 = static_cast<size_t>(::floor(seq[f]));
		size_t f2 = static_cast<size_t>(::ceil(seq[f]));
		float t = f1 == f2 ? 0 : static_cast<float>(seq[f] - ::floor(seq[f])) / static_cast<float>(f2 - f1);

		D3DXVECTOR3 v1 = src.GetPosition(f1), v2 = src.GetPosition(f2);
		res.SetPosition(f, v1 * (1.0f - t) + t * v2);
		for (size_t j = 0;j < src.NumJoints();j++)
		{
			D3DXQUATERNION q, q1 = src.GetRotation(f1, j), q2 = src.GetRotation(f2, j);
			D3DXQuaternionSlerp(&q, &q1, &q2, t);
			res.SetRotation(f, j, q);
		}
	}

	return res;
}
