float4x4 g_mWorld : World;
float4x4 g_mWorldIT : WORLDINVERSETRANSPOSE;
float4 g_vEyePos : EyePos;
float4x4 g_mWorldViewProj : WorldViewProj;

const float4 g_vLightPos = float4(-10.0f, 20.0f, 0.0f, 1.0f);
const float4 g_vLightColor = float4(0.8f, 0.8f, 0.6f, 1.0f);
const float g_fSpecExpon = 0.7f;

float4 g_vAmbiColor =float4(0.8f, 0.8f, 0.8f, 1.0f);
float4 g_vSurfColor =float4(0.8f, 0.8f, 0.8f, 1.0f);

struct PHONG_VS_INPUT
{
  float4 vPos : POSITION;
  float3 vNormal : NORMAL;
  float2 vUV : TEXCOORD0;
};

struct PHONG_VS_OUTPUT
{
  float4 vPos : POSITION;
  float4 Color : COLOR;
};

PHONG_VS_OUTPUT PhongVS(PHONG_VS_INPUT In)
{
  PHONG_VS_OUTPUT Out = (PHONG_VS_OUTPUT)0;

  float3 Pw = mul(In.vPos, g_mWorld);
  float3 Nn = mul(In.vNormal, g_mWorldIT);
  Nn = normalize(Nn);
  
  float3 Vn = normalize(g_vEyePos - Pw);
  float3 Ln = normalize(g_vLightPos - Pw);
  float3 Hn = normalize(Vn + Ln);

  Out.vPos = mul(In.vPos, g_mWorldViewProj);

  float4 litV = lit(dot(Nn, Ln), dot(Nn, Hn), g_fSpecExpon);
  float4 diffContrib = g_vSurfColor * ((0.5 * litV.y * g_vLightColor) + g_vAmbiColor);
  float4 specContrib = litV.y * litV.z * g_vLightColor;
  Out.Color = diffContrib + specContrib;
  
  return Out;
}

technique PhongShader
{
  pass P0
  {
    VertexShader = compile vs_1_1 PhongVS();
    PixelShader = NULL;
  }
}
